
:- ensure_loaded( library(lists) ).		% memberchk/2, member/2.
:- ensure_loaded( '../src/init_lib' ).       % library/1, requires/1.

:- requires( opts_select_infiles/3 ).
:- requires( werr/1 ).
:- requires( has_extension/3 ).
:- pl( swi(_), 
          requires( fget_line_swi/2 ),
          requires(fget_line/2) ).

comma_delim_to_data( Opts ) :-
	opts_select_infiles( Opts, Files, NFlOpts ),
	( memberchk(class(ClassPos),NFlOpts) ->
		true
		;
		werr( [['No class/1 option was given. Aborting...'],[]] ),
		abort
	),
	findall( Ign, member(ign(Ign),NFlOpts), Igns ),
	sort( Igns, SrIgns ),
	adjust_ignores_that_follow_class_pos( SrIgns, ClassPos, AdjIgns ),
	comma_delim_files_to_data( Files, ClassPos, AdjIgns ).

comma_delim_files_to_data( [], _CPos, _Igns ).
comma_delim_files_to_data( [F|Fs], CPos, Igns ) :-
	open( F, read, In ),
	has_extension( F, Stem, _Ext ),
	atom_concat( Stem, '_data.pl', DataF ),
	open( DataF, write, Out ),
	comma_delim_stream_to_data( In, Out, CPos, Igns, _NofAttrs ),
	close( In ),
	close( Out ),
	comma_delim_files_to_data( Fs, CPos, Igns ).

comma_delim_stream_to_data( In, _Out, _CPos, _Igns, _NofAttrs ) :-
	at_end_of_stream( In ),
	!.
comma_delim_stream_to_data( In, Out, CPos, Igns, NofAttrs ) :-
	fget_line( In, Line ),
	read_line_to_term( Line, CPos, Igns, NofAttrs, Term ),
	writeq( Out, Term ), write( Out, '.' ), nl( Out ),
	comma_delim_stream_to_data( In, Out, CPos, Igns, NofAttrs ).

read_line_to_term( Line, CPos, Igns, Arity, Term ) :-
	comma_delim_string_of_nums_or_atoms( Line, NumsOrAtoms ),
	nth( CPos, NumsOrAtoms, Class, NorARed ),
	nths( Igns, NorARed, _Nths, Attrs ),
	( length(Attrs,Arity) -> true ;
		werr( [[mismatch_in_nums_of_attributes(length(Attrs,Arity))]] ),
		abort
	),
	append( Attrs, [Class], Args ),
	Term =.. [data|Args].

nths( [], Red, [], Red ).
nths( [N|Ns], List, [Nth|Nths], Red ) :-
	nth( N, List, Nth, RemList ),
	nths( Ns, RemList, Nths, Red ).

comma_delim_string_of_nums_or_atoms( Line, Atomics ) :-
	append( LeftString, [0',|RightString], Line ),
	!,
	( catch(number_codes(Atomic,LeftString),_Any,fail) ->
		true
		;
		atom_codes( Atomic, LeftString )
	),
	Atomics = [Atomic|TAtomics],
	comma_delim_string_of_nums_or_atoms( RightString, TAtomics ).
comma_delim_string_of_nums_or_atoms( Line, [Atomic] ) :-
	( catch(number_codes(Atomic,Line),_Any,fail) ->
		true
		;
		atom_codes( Atomic, Line )
	).

% adjust_ignores_that_follow_class_pos( Igns, CPos, Adjusted ) :-
% All Igns that follow CPos, have 1 subtracted from them, since CPos
% will be the first removed from any lines read.
%
adjust_ignores_that_follow_class_pos( [], _CPos, [] ).
adjust_ignores_that_follow_class_pos( [H|T], CPos, [AdjH|AdjT] ) :-
	( H < CPos -> AdjH is H ; AdjH is H - 1 ),
	adjust_ignores_that_follow_class_pos( T, CPos, AdjT ).
