
:- ensure_loaded( '../../src/init_lib' ).

:- requires( opts_have_help/1 ).
:- requires( opts_select_infiles/3 ).
:- requires( select_n_uniformly/5 ).
:- requires( write_terms/4 ).
:- requires( n_distinct_randoms_in/4 ).
:- requires( split_on_indices/4 ).

pare_test_set( Opts ) :-
	opts_have_help( Opts ),
	!,
     shell( clear ), nl, nl,
	write( 'pare_test_set:      pare a part from a dataset to act as a test-set.' ), nl, nl,
	write( '  Options.' ), nl,
	write( '      arity(A),     arity of the data/n clause' ), nl,
     write( '      file(D),      the datafile to pare (atomic also)' ), nl,
	write( '      method(M)     method to use: *ord* or rnd.' ), nl,
	write( '      part(P)       a percentage, fraction or expression' ), nl,
     nl, nl.

pare_test_set( Opts ) :-
	\+ memberchk( part(_), Opts ),
	!,
	write( user_error, option_required(part/1) ), nl( user_error ).
pare_test_set( Opts ) :-
	\+ memberchk( arity(_), Opts ),
	!,
	write( user_error, option_required(arity/1) ), nl( user_error ).

pare_test_set( InOpts ) :-
	% memberchk( datafile(D), Opts ),
	opts_select_infiles( InOpts, Files, Opts ),
	( Files == [] -> 
		write( user_error, option_required(datafile/1) ), nl( user_error ),
		abort
		;
		Files = [D|RstD],
		( RstD == [] -> 
			true 
			;
			write( user_error, ignoring_files(RstD) ), nl( user_error )
		)
	),
	ensure_loaded( D ),
	memberchk( arity(A), Opts ),
	functor( Datum, data, A ),
	findall( Datum, Datum, Data ),
	length( Data, NofDpnts ),
	memberchk( part(P), Opts ),
	partition_as_abs_of_population( P, NofDpnts, AbsToRmv ),
	( memberchk( method(Method), Opts ) -> true; Method=ord ), 
	method_on_data( Method, AbsToRmv, Data, NofDpnts, Sel, LeftOver ),
	pare_datafile( D, TrainF, TestF ),
	write_terms( TrainF, write, LeftOver, [quoted(true)] ),
	write_terms( TestF, write, Sel, [quoted(true)] ).

method_on_data( rnd, AbsToRmv, Data, NofDpnts, Sel, LeftOver ) :-
	select_n_uniformly( AbsToRmv, Data, NofDpnts, Sel, LeftOver ).
method_on_data( ord, N, Data, DataPnts, Sel, LeftOver ) :-
	Upper is DataPnts + 1,
	n_distinct_randoms_in( N, 1, Upper, Rnds ),
	split_on_indices( Data, Rnds, Sel, LeftOver ).

partition_as_abs_of_population( Part, Pop, Abs ) :-
	% allow for P = 5%, useful when 0.8% is desired.
	P is Part,
	( P < 1 -> 
		Abs is integer(Pop * P)
		;
		Abs is integer((Pop * P) / 100)
	).

pare_datafile( InFile, Train, Test ) :-
	( atom_concat(Stem,'.pl',InFile) ->
		true
		;
		Stem = InFile
	),
	atom_concat( Stem, '_train', TrainStem ),
	atom_concat( Stem, '_test', TestStem ),
	atom_concat( TrainStem, '.pl', Train ),
	atom_concat( TestStem, '.pl', Test ).
