:- ensure_loaded( library(lists) ).		% memberchk/2, member/2.
:- ensure_loaded( library(charsio) ).		% write_to_chars/2.

:- ensure_loaded( '../auxil/n_samples' ).		% /3.
:- ensure_loaded( 'eps_n_latex_graph_freqs' ).	% /7.
:- ensure_loaded( '../src/init_lib' ).
:- ensure_loaded( library(slp_in_options) ).		% /2.
:- ensure_loaded( library(tmp_in_options) ).		% /3.
:- ensure_loaded( library(werr) ).				% /1.
:- ensure_loaded( library(is_list_of_n_vars) ).	% /2.
:- ensure_loaded( library(kv_transpose) ).		% /2.
:- ensure_loaded( library(write_list_elem_per_line_w_period) ).	% /1.

/*
:- ensure_loaded( n_samples_tex ).

part_bn_test :- n_samples_tex( [
				call( bn([1,2,3,4],[1-[],2-[1],Third,4-[1]]), Third ),
				out( part_bn_out ),
				samples( 10000 ),
				slp( '../bns/slps/bn_or_any_pick_rcp_6h34' ),
				tmp( part_bn_tmp )
	] ).
     */

/* 
upsh n_samples_tex call=pedigree=Ped:Ped stem=ped_ns_eff samples=25000 slp=weighed_eff sort=times with=lines precall=ensure_loaded=tsar_persons
*/
n_samples_tex( Opts ) :-
	memberchk( help, Opts ),
	!, nl,
	write( 'Create a tex source for N samples drawn independently from an SLP.' ),
	nl, nl,
	write( '  call/2:              Callable and Observable'), nl,
	write( '  col/1:               number of collumns in latex'), nl,
	write( '  dot_ith_intervals/1: number of completion intervals at which a dot will be printed' ), nl,
	write( '  gv/1:                boolean, use gv to display .ps file. Default is false' ), nl,
	write( '  out/1:               stem of filename for tex and ps output'), nl,
	write( '  precall/1:           a call invoked once, just before the sampling starts' ), nl,
	write( '  post/1:              a call invoked once, immediately after the sampling finishes' ), nl,
	write( '  samples/1:           number of samples to be drawn' ), nl,
	write( '  slp/1:               the SLP to use' ), nl,
	write( '  sort/1:              boolean (false), or times, if set sorts the frequencies before drawing the graph' ), nl,
	write( '  stem/1:              can be used instead of the combination of slp/1 and tmp/1' ), nl,
	write( '   terms/1,           boolean (true), record observations as terms.' ), nl,
	write( '  tmp/1:               temporary file to use, for holding transformed SLP when reading-in' ), nl, nl.

n_samples_tex( Opts ) :-
	( slp_in_options(Opts,Slp) -> true ; 
	 	werr( [['slp/1 option missing. Aborting.']] ),
	  	abort 
	),
	( memberchk(call(Call,Obs),Opts) -> true ;
	 	werr( [['call/2 option missing. Aborting.']] ),
	  	abort
	),
	( tmp_in_options(Opts,[],Tmp) -> true ;
		Tmp = n_samples_tex_tmp
	),
	( memberchk(samples(Samples),Opts) -> true ;
		werr( [['Number of samples to draw was no specified (samples/1).'],
			  ['Will draw 1000 samples.']] ),
		Samples is 1000
	),
	( memberchk(precall(PreC),Opts) -> true ;
		PreC = true ),
	( memberchk(dot_ith_intervals(Dii),Opts) -> 
		Topts = [dot_ith_intervals(Dii)]
		;
		Topts = []
	),
	n_samples( [samples(Samples),call(Call-Obs),slp(Slp),tmp(Tmp),precall(PreC)|Topts], TmpNS, PrvFrqs ),

	( memberchk(sort(true),Opts) -> 
		keysort(PrvFrqs,Frqs) 
		; 
		( memberchk(sort(times),Opts) -> 
			kv_transpose( PrvFrqs, TrspPrvFrqs ),
			keysort(TrspPrvFrqs, TrspSrtFrqs ),
			kv_transpose( TrspSrtFrqs, Frqs )
			% reverse( RevFrqs, Frqs )
			;
			PrvFrqs = Frqs
	  	)
	),
	write( tmp(Tmp)-tmpns(TmpNS) ), nl,
	( memberchk(out(Out),Opts) -> true ;
		( memberchk(stem(Stem),Opts) -> 
			atom_concat( Stem, '_samples', Out )
			; 
			Out = 'samples'
		)
	),
	( memberchk(col(Cols),Opts) -> true ;
		Cols is 2 ),
	number_codes( Samples, SamplesCs ),
	atom_codes( SamplesAtm, SamplesCs ),
	write_to_chars( Call, CallCs ),
	length( CallCs, CallLgt ),
	( CallLgt > 100 -> CallAtm = '_long_' ; atom_codes( CallAtm, CallCs ) ),
	write_to_chars( Obs, ObsCs ),
	atom_codes( ObsAtm, ObsCs ),
	flatoms( ['\\verb&Frequencies of ',SamplesAtm,' samples drawn independently from program&\n\\verb& ',Slp,' using goal ',CallAtm,' and observing ',ObsAtm,'.&'], For ),
	select_latex_opts( Opts, LtxOpts ),
	eps_n_latex_graph_freqs( Frqs, Out, For, Cols, Slp, TmpNS, LtxOpts ),
	( memberchk(terms(false),Opts) -> 
		true
		;
		atom_concat( Out, '.obs', ObsFile ),
		tell( ObsFile ),
		write_list_elem_per_line_w_period( Frqs ),
		told
	),
	delete_file( TmpNS ).

select_latex_opts( Opts, LtxOpts ) :-
	findall( O, 
				( member(Name/Ar,[gv/1,with/1]),
				  is_list_of_n_vars(Ar,Args),
				  O =.. [Name|Args],
				  memberchk( O, Opts )
				),
			LtxOpts ).
