:- ensure_loaded( 'src/init_lib' ).

:- ensure_loaded( library(pl) ). 		% pl/2.
:- ensure_loaded( library(lists) ). 	% is_list/1.

:- pl( yap(_Y), set_prolog_flag(update_semantics,logical) ).
:- pl( yap(_Y), set_prolog_flag(unknown,error) ).
:- pl( yap(_Y), set_prolog_flag(single_var_warnings,on) ).

:- ensure_loaded( 'src/all_dynamic' ).
:- ensure_loaded( 'src/ad_to_slp' ).
:- ensure_loaded( 'src/kernel' ).
:- ensure_loaded( 'src/version' ).
:- elm( 'src/model_type_calls' ).

:- ensure_loaded( library(lss) ).					% lss/0.
:- ensure_loaded( library(flatoms) ).				% /2.
% :- ensure_loaded( library(unique_filename) ).		% /2.
:- ensure_loaded( library(random) ).
:- ensure_loaded( library(copy_stream_on_stream) ).
:- ensure_loaded( library(dbg) ).					% dbg/1,2.
:- ensure_loaded( library(werr) ).					% /1.
:- ensure_loaded( library(error_to_output) ).		% /2.
:- ensure_loaded( library(head_to_spec) ).			% /2.
:- ensure_loaded( library(esr) ).					% /1.
:- ensure_loaded( library(mcmcms_top_dir) ).			% bb_puts it.
:- requires( elm/1 ).
:- requires( is_list_of_n_vars/2 ).
:- requires( to_list/2 ).
:- requires( unique_filename/3 ).
:- requires( is_letter/1 ).

:- pl( sicstus(_S), ensure_loaded( library(number_atom) ) ).  		% /2.

% Args = 
% Out, Stats, Kernel, ModelType, Prior, PB, Repeats, Data, HotChainsIds, Seeds1-3, SGl.

/*
     F: 
     S: 
     K: 
     M: 
     B: 
     P: 
     R:
     D: 
     H:
     S1-3: random seeds
     Goal: 

*/
mcmcms( F,S,K,M,B,P,R,D,H,S1,S2,S3,SGl ) :-
	Np = P,
	number_atom_list( [R,S1,S2,S3], [Nr,Ns1,Ns2,Ns3] ),
	esr(open( S, append, Stats )),
	( var(Stats) -> Stats = user ; true ),
	esr(open( F, write, Results )),
     ( model_type_calls( M ) -> true
          ; werr( failure_of_model_calls_for(M), error ) ),
	print_args( Stats, F,S,K,M,B,P,R,D,H,S1,S2,S3 ),
	load_data_and_put_dgoal( D ),
	setrand( rand(Ns1,Ns2,Ns3) ),
	( bb_get(msd,_) -> true ; bb_put(msd,rm) ),
	bb_put( bk_failures, [] ),
	all_dynamic( B ),
     UFOpts = [report(false),add(environ('HOSTNAME'),8)],
     unique_filename( translated__slp, UFOpts, TmpLoad ),
	ad_to_slp( [rm(false),tmp(TmpLoad)] ),
	( var(Stats) -> WSon = user ; WSon = Stats ),
     mcmcms_maintainer( Maintainer ),
	Failure = [['Main MCMC call ended with failure.'],
	           ['Check your SLP, the arguments of run_data/9 and read the manual.'],
			 ['If you cannot resolve the problem dont hasitate to send a bug report, to:'],
			 [Maintainer]],
	( main_out( WSon, Results, S, K, SGl, Nr, Np, H, TmpLoad ) ->
				true
			; 	werr( Failure ), abort
	),
	esr(close( Results )),
	%% esr(set_prolog_flag( user_error, OldError )),
	esr(close( Stats )).

main_out( Stats, Results, S, Kid, SGl, Nr, Np, H ) :-
	main_out( Stats, Results, S, Kid, SGl, Nr, Np, H, none__ ).

main_out( Stats, Results, S, Kid, SGl, Nr, Np, H, TmpLoad ) :-
	%% esr(set_output( Stats )),
	% for the following, use pl/1 instead/additionally.
	( TmpLoad == none__ ->
		true
		;
		open( TmpLoad, read, ReadFrom  ),
          ( atom_concat(StatsStem,'.stats',S) ->
               atom_concat( StatsStem, '.dlp', SlpF ),
               open( SlpF, write, SlpO )
               ;
		     Stats = SlpO
          ),
		copy_stream_on_stream( ReadFrom, SlpO ),
		close( ReadFrom ),
		current_output( OldCOut ),
		werr( ['--- start listing of last_skips/2'], info, '', SlpO ),
		set_output( SlpO),
		listing( last_skips/2 ),
		set_output( OldCOut ),
		werr( ['--- end   listing of last_skips/2'], info, '', SlpO ),
          ( var(SlpF) -> true; close(SlpO) ),
		delete_file_if( TmpLoad )
	),
	main_statistics( StartAt, Stats ),
	esr( current_output(OldOut) ),
	esr( set_output(Results)),
	%% bid_exclude_lbl( Kid ),
	%%% testing Swi's randoms.
	%%% getrand( Got ),
	%%% random( R ),
	%%% write( user_error, Got-R ), abort,
	kernel( SGl, H, Nr, Kid, Np ),
	%% esr(set_output( Stats )),
	esr( set_output(OldOut) ),
	main_statistics( EndWith, Stats ),
	Total is (EndWith - StartAt) / 1000,
	% write( Stats, ((EndWith - StartAt) / 1000) ), write( Stats, ' = ' ), nl(Stats),
	% writeq( Stats, Total ), nl(Stats),
     nl( Stats ),
     pc( Stats, total(((EndWith - StartAt) / 1000), Total) ),
	bb_get( bk_failures, Failures ),
	( Failures = [] -> 
		% write( Stats, 'There were no backtracking failures.' )
		% just be silent...
		true
		;
		write( Stats, 'Backtracking failed on the following iterations: ' ),
		write( Stats, Failures )
	),
	nl( Stats ).

print_args( Stats, F,S,K,M,B,P,R,D,H,S1,S2,S3 ) :-
	mcmcms_version( MsVer ),
     host_name( Host ),
	prolog_flag( version, ThisPlEngine ),
	pc( Stats, on(ThisPlEngine) ),
     (bb_get(last_data, ld(_TheData,DataFile)) ->
          % a hack communication with run.pl
          true
          ;
          DataFile = D % and hope for the best
     ),
	pc( Stats, [mcmcms(MsVer),host(Host)] ),
     nl( Stats ),
     % functor( D, NameD, ArD ),
     pc( Stats, [model_type(M),results_file(F),stats_file(S),
		  kernel(K),prior_file(B), backtrack_probability(P),
		  repeats(R),data_call(D), data_file(DataFile), hot_chains(H),
            seeds(S1,S2,S3)] ),
     nl( Stats ).
           

number_atom_list( [], [] ).
number_atom_list( [H|T], [Nh|Nt] ) :-
	( number(H) ->
		Nh = H
		;
		number_atom( Nh, H )
	),
	number_atom_list( T, Nt ).

/*
bn_file_abs( Bn, AbsBn ) :-
	( pl(yap(_Y)) -> 
		library_directory( Lib ),
		flatoms( [Lib,'/',Bn], AbsBn ),
		file_exists( AbsBn ), 
		!
		;
		AbsBn = library( Bn )
	).
*/

main_statistics( Sum, Stats ) :-
	datime( Datime ), 
     nl( Stats ),
     pc( Stats, Datime ),
	prolog_flag( user_error, Old, Stats ),
	pl( Pl ),
	( Pl = sicstus(_) ->
		statistics( atom_garbage_collection, [_Nofag,_Ab,OvAt] )
		;
		( Pl = yap(_) ->
			statistics( cputime, [_OvCt,CpuTime] ),
			pc( Stats, yap_cputime(CpuTime) ),
			OvAt is 0
			; % else it is swi
			statistics( cputime, CpuTime ),
			pc( user_error, swi_cputime(CpuTime) ),
			OvAt is 0
		)
	),
	statistics( runtime, [OvRt,_RlRt] ),
	statistics( garbage_collection, [_Nofgc,_Bfrd,OvGt] ),
	( statistics( stack_shifts, [_Nofls,_Nofts,OvSt] ) ->
		true; OvSt is 0 ),
	Sum is (OvRt + OvGt + OvSt + OvAt),
	pl( swi(_), 
		( current_output(SwiOut),
		  set_output(Stats),
            nl, write( '/*' ), nl,
		  statistics,
            nl, write( '*/' ), nl,
		  set_output(SwiOut)
		)
		,
          ( nl(Stats), write(Stats,'/*'),nl(Stats),
            statistics, write(Stats, '*/' ),nl(Stats))
	),
	prolog_flag( user_error, _St, Old ).

load_data_and_put_dgoal( D ) :-
	head_to_spec( D, SpecD ),
	( current_predicate(user:SpecD) -> 
		( call( user:D ) ->
			true
			;
			werr( [['Warning: data call failure, for: ', D ]] )
		)
		;
		( D == null ->
			true
			;
			werr( [['Warning: data call does not exists, for: ', D ]] )
		)
	),
	% findall( DA, current_predicate(data/DA), DAs ),
	% Change 2.
	( current_predicate(data/2) ->
		( (data(Instance,_InstTms),is_list(Instance)) -> 
			length( Instance, InstLength ),
			is_list_of_n_vars( InstLength, VarListOfLengthVs ),
			bb_put( multi_data_format, data(VarListOfLengthVs,_TmsVar) )
			;
			werr( [['Cannot locate data/2. Aborting...']] ), abort
		)
		;
		true
	).
	/* 
	( DAs = [OkDA] -> 
		( OkDA = 
		functor( Datum, data, OkDA ),
		bb_put( data_format, Datum ),
		MultiDA is OkDA + 1,
		functor( MultiDatum, data, MultiDA ),
		bb_put( multi_data_format, MultiDA-MultiDatum )
		;
		werr( [['There should be only one data/n predicate defined.'],
		       ['Aborting, after finding multiple definitions: ',DAs]] ),
			  abort
	).
	*/

pc( Stream, Clause ) :-
     to_list( Clause, Clauses ),
     pc_1( Clauses, Stream ).

pc_1( [], _Stream ).
pc_1( [H|T], Stream ) :-
     portray_clause( Stream, H ),
     pc_1( T, Stream ).
