:- ensure_loaded( '../../src/init_lib' ).
:- ensure_loaded( library(directory_files_regular) ).	% /2.
:- ensure_loaded( library(sys_info) ).				% /1.
:- ensure_loaded( library(lists) ).				% append/3, select/3.

gz_comparison( Opts ) :-
	write( opts(Opts) ), nl,
	( select(mode(Mode),Opts,Dirs) -> true ; Mode=gzip, Dirs = Opts ),
	write( dirs(Dirs) ), nl,
	gz_comparison_1( Dirs ).

gz_comparison_1( [] ).
gz_comparison_1( [Dir|TDirs] ) :-
	working_directory( Old, Dir ),
	directory_files_regular( data, DataFiles ),
	working_directory( FDir, data ),
	gz_data( DataFiles, Mode ),

	working_directory( _, FDir ),
	directory_files_regular( losses, LossFiles ),
	working_directory( _FDir1, losses ),
	gz_losses( LossFiles, Mode ),

	working_directory( _, FDir ),
	directory_files_regular( models, ModelFiles ),
	working_directory( _FDir2, models ),
	gz_models( ModelFiles, Mode ),
	working_directory( _, Old ),
	gz_comparison_1( TDirs ).

gz_data( [], _ ).
gz_data( [H|T], Mode ) :-
	( Mode == gunzip -> 
		( atom_concat(_Stem,'.gz',H) ->
			atom_concat( 'gunzip ', H, Gunzip ),
			sys_info( Gunzip )
			;
			true
		)
		;
		( atom_concat(_Stem,'.gz',H) ->
			true
			;
			atom_concat( 'gzip ', H, Gzip ),
			sys_info( Gzip )
		)
	),
	gz_data( T, Mode ).

gz_losses( [], _Mode ).
gz_losses( [H|T], Mode ) :-
	( Mode == gunzip ->
		( atom_concat(_Stem,'.gz',H) ->
			atom_concat( 'gunzip ', H, Gunzip ),
			sys_info( Gunzip )
			; 
			true
		)
		;
		atom_codes( H, HCs ),
		( append(_Stem,[0'.,0'l,_,_],HCs) ->
			atom_concat( 'gzip ', H, Gzip ),
			sys_info( Gzip )
			;
			true
		)
	),
	gz_losses( T, Mode ).

gz_models( [], _Mode ).
gz_models( [H|T], Mode ) :-
	( Mode == gunzip -> 
		( atom_concat(_Stem,'.gz',H) ->
			atom_concat( 'gunzip ', H, Gunzip ),
			sys_info( Gunzip )
			;
			true
		)
		;
		( (atom_concat(_Stem,'.gz',H);atom_concat(_Stem,'.stats',H)) ->
			true
			;
			atom_concat( 'gzip ', H, Gzip ),
			sys_info( Gzip )
		)
	),
	gz_models( T, Mode ).
