:- ensure_loaded( library(lists) ).		% memberchk/2, member/2, append/3.
:- ensure_loaded( library(system) ).		% /1, file_exists/1.

:- ( file_exists( expected_loss) -> ensure_loaded( expected_loss )
     ; ensure_loaded( '../expected_loss' ) ). 		% /3.

post_loss( Opts ) :-
	( Opts == help ; memberchk(help,Opts) ),
	!,
	write( 
'post_loss/1       post processes files that contain MCMCMS samples
                   to produce the loss incurred by each sample according
                   to a set of loss functions. Two files are generated per
                   loss function given. One is the list of all individual 
                   losses, while the other contains a single term with
                   the expected loss of the sample. Note that you should
                   make sure any predicates such as true_bn/1 and true_eg/1
                   should be in memory (e.g. with appropriate us of load/1.
                   The following options are allowed: 
         
         compress( Bool )   whether to use gzip on the output. Def. is true.
         dir( Dir ),        directory for the resulting files. Only one should 
                            be given. When not present, \'.\' is used.
         mcmc(MCMC),        chain file containing the MCMCMS samples.
         load( Fname ),     filename to be loaded before calculating the losses.
         loss( Pname, ID ), Pname is the predicate name of <loss>(Model,Loss), 
                            and ID is a two codes abbreviation for the function.
                            User must ensure Pname/2 is defined, e.g. in one
                            of the files given in load/1.
         out( -Files )      the loss output files. (Not expected loss.)

Note that multiple gz/1, load/1 and loss/2 are allowed.
Example: upsh post_loss.pl load=loss loss=loss_ck,ck loss=loss_ks,ks mcmc=atm=models/cm_tts_d100_mrm_buc_r100_pts_hua_s1.gz load=simple dir=losses
where simple.pl contains,
true_eg([1-4,2-4,3-4]).
true_bn([1-4,2-4,3-4]).

The above works fine in the directory created by the simple example
in comparison/comparison.pl .

' ).

post_loss( Opts ) :-
	( memberchk(dir(Dir),Opts) -> true ; Dir = '.'),
	findall( Ld, (member(load(Ld),Opts),ensure_loaded(Ld)), _Lds ),
	findall( L, (member(loss(L,Lid),Opts),assert(loss_id(L,Lid)) ), Ls ),
	findall( F, member(mcmc(F),Opts), Fs ),
	( memberchk(out(AllOutFs),Opts) -> true ; true ),
	post_files( Fs, Ls, Dir, [], AllOutFs ).

post_files( [], _Ls, _Dir, OutFs, OutFs ).
post_files( [F|Fs], Ls, Dir, AccOutFs, OutFs ) :-
	atom_concat( 'gunzip ', F, Gunzip ), 
	system( Gunzip ),
	atom_concat( StmF, '.gz', F ),
	expected_loss( Ls, Dir, StmF, LsOutFs ),
	atom_concat( 'gzip ', StmF, Gzip ),
	system( Gzip ),
	append( LsOutFs, AccOutFs, NxOutFs ),
	post_files( Fs, Ls, Dir, NxOutFs, OutFs ).
