:- ensure_loaded( library(lists) ).		% reverse/2.
:- ensure_loaded( library(ordsets) ).	% list_to_ord_set/2.
:- ensure_loaded( '../../../src/init_lib' ).
:- ensure_loaded( library(file_to_list_of_chars) ). % stream_to_list_of_chars/2

test :- bif2pl( 'alarm.bif' ).

bif_to_slp( [] ).
bif_to_slp( [F|Fs] ) :-
	bif_to_pl( F, Str ),
	atom_concat( Stem, '.bif', F ), 
	atom_concat( Stem, '.slp', S ),
	atom_concat( Stem, '_vrs.pl', V ),
	open( S, write, Out ),
	open( V, write, Vrs ),
	write( Vrs, ':- discontiguous domain_size/2, possible_values/2.' ),
	nl( Vrs ), nl( Vrs ),
	plbn_writes_slp( Str, Out, Vrs ),
	write( ok(F) ), nl,
	close( Out ), close( Vrs ),
	bif_to_slp( Fs ).
	
plbn_writes_slp( rbn(_Nm,_Ats,Vs,Pts), O, Vrs ) :-
	write( O, '% Automaticaly tranlated from .bif format' ), nl( O ),
	write( O, '% ' ), nl( O ), nl( O ),
	bn_vars_paired( Vs, 1, O, Pts, VPrs ),
	order_vars( VPrs, [], 0, [], OrdPrs, SecPrs ),
	write( O, 'sample( [' ), nl( O ),
	write_sample_head_list_of_vars( SecPrs, O ),
	write( O, '                  ] ) :-'  ), nl( O ),
	sample_clause_body( OrdPrs, O ),
	p_clauses( VPrs, Pts, O ),
	vars_dmn_preds( OrdPrs, 1, Vrs ).

order_vars( [], Acc, Cnt, Pspnd, OrdPrs, SecPrs ) :-
	( Pspnd == [] -> 
		reverse( Acc, OrdPrs ),
		% section_vars_on_parents_length( OrdPrs, 0, SecPrs )
		% use the above if you need sectioning.
		SecPrs = OrdPrs
		;
		( Cnt =:= 0 ->
			write( user_error, cannot_order_these(Pspnd) ), nl( user_error ),
			abort
			;
			order_vars( Pspnd, Acc, 0, [], OrdPrs, SecPrs )
		)
	).

order_vars( [V-N-Qs-Dmn|T], Acc, Cnt, Pspnd, OrdPrs, SecPrs  ) :-
	( all_qs_present(Qs,Acc) -> 
		NxAcc = [V-N-Qs-Dmn|Acc],
		NxCnt is Cnt + 1,
		NxPspnd = Pspnd
		;
		NxAcc = Acc, NxCnt is Cnt,
		NxPspnd = [V-N-Qs-Dmn|Pspnd]
	),
	order_vars( T, NxAcc, NxCnt, NxPspnd, OrdPrs, SecPrs ).

section_vars_on_parents_length( [], _N, Sec ) :-
	!,
	Sec = [].
section_vars_on_parents_length( Ord, N, Sec ) :-
	n_section_vars_on_parents_length( Ord, N, Sec, Rem, TSec ),
	NxN is N + 1,
	section_vars_on_parents_length( Rem, NxN, TSec ).

n_section_vars_on_parents_length( [], _N, Sec, [], Sec ).
n_section_vars_on_parents_length( [V-NodeN-Qs-Dmn|T], N, Sec, Rem, CntSec ) :-
	( length(Qs,N) -> 
		Sec = [V-NodeN-Qs-Dmn|TSec],
		Rem = TRem
		;
		Sec = TSec,
		Rem = [V-NodeN-Qs-Dmn|TRem]
	),
	n_section_vars_on_parents_length( T, N, TSec, TRem, CntSec ).

all_qs_present( [], _Acc ).
all_qs_present( [H|T], Acc ) :-
	memberchk( H-_N-_HQs-_HDmn, Acc ),
	all_qs_present( T, Acc ).

bn_vars_paired( [], _, _, _, [] ).
bn_vars_paired( [v(V,_Sz,Dmn)|T], N, O, Pts, [V-N-QVs-Dmn|Prs] ) :-
	NxN is N + 1,
	memberchk( t(V,QVs,_), Pts ),
	bn_vars_paired( T, NxN, O, Pts, Prs ).

write_sample_head_list_of_vars( [], _O ).
write_sample_head_list_of_vars( [V-_-_-_|T], O ) :-
	write( O, '            ' ), write( O, V ),
	( T == [] -> true ; write( O, ',' ) ),
	nl( O ),
	write_sample_head_list_of_vars( T, O ).

vars_dmn_preds( [], Nvrs, VOut ) :-
	Rnvrs is Nvrs - 1,
	nl( VOut ), write( VOut, num_rvs( Rnvrs ) ),
	write( VOut, '.' ), nl( VOut ).
vars_dmn_preds( [_V-_-_-Dmn|T], Nvrs, VOut ) :-
	list_to_ord_set( Dmn, OrdDmn ),
	writeq( VOut, possible_values(Nvrs,OrdDmn) ), 
	write( VOut, '.' ), nl( VOut ),
	length( Dmn, DmnSz ),
	writeq( VOut, domain_size(Nvrs,DmnSz) ), 
	write( VOut, '.' ), nl( VOut ), nl( VOut ),
	NxNvrs is Nvrs + 1,
	vars_dmn_preds( T, NxNvrs, VOut ).

p_clauses( [], _Pts, _O ).
p_clauses( [V-N-_Qvs-Dmn|T], Pts, O ) :-
	nl( O ),
	memberchk( t(V,_CdVs,Tbl), Pts ), 
	conditioned_clause( Tbl, N, Dmn, O ),
	p_clauses( T, Pts, O ).

conditioned_clause( [], _N, _Dmn, _O ).
conditioned_clause( [e(CdVls,Prbs)|T], N, Dmn, O ) :-
	single_conditioned( Prbs, Dmn, N, CdVls, O ),
	( T == [] -> true ; write( O, '%' ), nl( O ) ),
	conditioned_clause( T, N, Dmn, O ).

single_conditioned( [], _, _N, _CdVls, _O ).
single_conditioned( [P|Ps], [E|Es], N, CdVls, O ) :-
	write( O, P ), 
	write( O, ' :: p_' ),
	% write( O, ' :: p( ' ),
	write( O, N ), write( O, '( ' ), writeq( O, E ),
	% write( O, N ), write( O, ', ' ), writeq( O, E ),
	write( O, ', ' ), writeq( O, CdVls ),
	write( O, ' ).' ), nl( O ),
	single_conditioned( Ps, Es, N, CdVls, O ).

bif_to_pl( File, Str ) :-
	open( File, read, In ), 
	stream_to_list_of_chars( In, Codes ),
	close( In ),
	phrase( bif_to_pl_dcg(Str), Codes ).
	% write( parsed(Str) ), nl.

bif_to_pl_dcg( rbn(Nid,Ndt,Vs,Prbs) ) -->
		network_id( Nid, Ndt ),
		variables( Vs ),
		prbs( Prbs ).

network_id( Nid, [] ) --> 
		next_token( 'network' ), 
		next_token( Nid ),
		next_token( '{' ),
		next_token( '}' ).

variables( [V|Vs] ) -->
	variable( V ), 
	!,
	variables( Vs ).
variables( [] ) --> {true}.

variable( v(Vname,NoV,Vals) ) --> 
		next_token( 'variable' ),
		next_token( Vname ),
		next_token( '{' ),
		next_token( type ),
		next_token( discrete ),
		next_token( '[' ),
		next_token( NoVAtm ),
		{ token_to_number( NoVAtm, NoV ) },
		next_token( ']' ),
		possible_token( '=' ),
		next_token( '{' ),
		comma_sep_to_list( Vals ),
		next_token( '}' ),
		% next_token( ';' ).
		next_token( '}' ).

prbs( [V|Vs] ) -->
	table( V ), 
	!,
	prbs( Vs ).
prbs( [] ) --> {true}.

table( t(Vid,CdVs,Tbl) ) -->
		next_token( 'probability' ),
		next_token( '(' ),
		next_token( Vid ),
		list_of_conditioning_vs( CdVs ),
		next_token( ')' ),
		next_token( '{' ),
		table_entries( Tbl ),
		next_token( '}' ).

%%%

next_token( Tok ) --> 
	single_code_token( Tok ), 
	!,
	consume_all_delimiters.
next_token( Tok ) --> next_token([],Tok,_).

next_token( Acc, Tok, Del ) --> 
	consume_delimiter( Del ), 
	!,
	consume_all_delimiters,
	{reverse(Acc,Rev),atom_codes(Tok,Rev)}.
next_token( Acc, Tok, Del ) --> 
	[C], {C \== 0')}, !,
	next_token( [C|Acc], Tok, Del ).
next_token( Acc, Tok, ')' ) --> 
	{reverse(Acc,Rev),atom_codes(Tok,Rev)}.

consume_delimiter( ' ' ) --> " ", !.
consume_delimiter( '\t' ) --> "\t", !.
consume_delimiter( '\n' ) --> "\n", !.
consume_delimiter( ',' ) --> ",".
consume_delimiter( ';' ) --> ";".

single_code_token( '=' ) --> "=", !.
single_code_token( '(' ) --> "(", !.
single_code_token( ')' ) --> ")", !.
single_code_token( '{' ) --> "{", !.
single_code_token( '}' ) --> "}".

consume_all_delimiters -->
	consume_delimiter( _ ),
	!,
	consume_all_delimiters.
consume_all_delimiters --> {true}.

comma_sep_to_list( [H|T] ) -->
	next_token( [], H, Del ),
	comma_sep_to_list_rec( Del, T ).

list_of_conditioning_vs( [H|T] ) -->
	next_token( '|' ), 
	!,
	next_token( [], H, Del ),
	comma_sep_to_list_rec( Del, T ).
list_of_conditioning_vs( [] ) --> {true}.

table_entries( [e(CdVls,PVals)|T] ) -->
		next_token( '(' ), 
		comma_sep_to_list( CdVls ),
		next_token( ')' ),
		comma_sep_to_list( PValAtms ),
		{atms_to_numbers( PValAtms, PVals )},
		% HERE convert to numbers.
		% next_token( ';' ),
		!,
		table_entries( T ).
table_entries( [e([],PVals)] ) -->
		next_token( table ), 
		!,
		comma_sep_to_list( PVals ).
		% next_token( ';' ).
table_entries( [] ) --> {true}.

%%%

sample_clause_body( [], _O ).
sample_clause_body( [V-N-QVs-_Dmn|T], O ) :-
	write( O, '      p_' ), write( O, N ), write( O, '( ' ), 
	% write( O, '      p( ' ), write( O, N ), write( O, ', ' ), 
	write( O, V ), write( O, ', ' ), write( O, QVs ),
	write( O, ' )' ),
	( T == [] -> write( O, '.' ) ; write( O, ',' ) ),
	nl( O ),
	sample_clause_body( T, O ).

%%%

comma_sep_to_list_rec( Del, [H|T] ) -->
	{Del == ','},
	!,
	next_token( [], H, NxDel ),
	comma_sep_to_list_rec( NxDel, T ).
comma_sep_to_list_rec( _, [] ) --> {true}.

token_to_number( Atm, Num ) :-
	atom_codes( Atm, Codes ),
	number_codes( Num, Codes ).

atms_to_numbers( [], [] ).
atms_to_numbers( [H|T], [N|Ns] ) :-
	atom_codes( H, Cs ),
	number_codes( N, Cs ),
	atms_to_numbers( T, Ns ).

possible_token( Tok ) -->
	next_token( Tok ),
	!.
possible_token( _ ) --> {true}.
