:- ensure_loaded(library(lists)).
:- ensure_loaded(library(ordsets)).

test_bn([f1-[xx(2),xy(2)],m1-[xx(1),xy(1)],xx(1)-[xx(4),xy(3)],xx(2)-[],xx(3)-[],xx(4)-[],xy(1)-[xx(3),xy(2)],xy(2)-[],xy(3)-[]]).

reachable(BN,X,E,R) :-
	phase1(E,[],BN,A),
	phase2([X-up],[],[],BN,A,E,R).


%From Koller p38 of Chapter 3
phase1([],AIn,_BN,AIn).
phase1(LIn,AIn,BN,AOut) :-
	select(Y,LIn,LMid),             % maintains the order
	!,
	(
	  \+ memberchk(Y,AIn)
	->
	  member(Y-Parents,BN),
	  list_to_ord_set(Parents,Parentsset),
	  ord_union(Parentsset,LMid,LNew)
	;
	  LNew = LMid
	),
	ord_add_element(AIn,Y,AMid),
	phase1(LNew,AMid,BN,AOut).



/*---------------------------------------------------------------------
  Predicate: phase2(+LIn,+VIn,+RIn,+BN,+A,+E,-ROut)
       Args: 
   Succeeds: exactly once
   Comments: phase2 of Koller algorithm
---------------------------------------------------------------------*/


phase2([],_VIn,R,_BN,_A,_E,R) :-
	!.

% delete already visited ones from LIn
phase2(LIn,VIn,RIn,BN,A,E,ROut) :-
	select(Y-D,LIn,LMid),
	memberchk(Y-D,VIn),
	!,
	phase2(LMid,VIn,RIn,BN,A,E,ROut).
	
phase2(LIn,VIn,RIn,BN,A,E,ROut) :-
	select(Y-D,LIn,LMid),
	ord_add_element(RIn,Y,RNew),
	ord_add_element(VIn,Y-D,VNew),
	!,
	(
	  (D=up,\+ memberchk(Y,E))
	->
	  memberchk(Y-YsParents,BN),
	  findall(Z-up,member(Z,YsParents),Zups),
	  list_to_ord_set(Zups,Zupset),
	  ord_union(LMid,Zupset,LMid2),
	  findall(Z-down,(member(Z-Parents,BN),member(Y,Parents)),Zdowns),
	  list_to_ord_set(Zdowns,Zdownset),
	  ord_union(LMid2,Zdownset,LNew)
	;
	  (
	    D=down
	  ->
	    (
	      \+ memberchk(Y,E)
	    ->
	      findall(Z-down,(member(Z-Parents,BN),member(Y,Parents)),Zdowns),
	      list_to_ord_set(Zdowns,Zdownset),
	      ord_union(LMid,Zdownset,LMid2)
	    ;
	      LMid2=LMid
	    ),
	    (
	      memberchk(Y,A)
	    ->
	      memberchk(Y-YsParents,BN),
	      findall(Z-up,member(Z,YsParents),Zups),
	      list_to_ord_set(Zups,Zupset),
	      ord_union(LMid2,Zupset,LNew)
	    ;
	      LNew=LMid2
	    )
	  ;
	    LNew = LMid
	  )
	),
	phase2(LNew,VNew,RNew,BN,A,E,ROut).
