/*---------------------------------------------------------------------
factors are like

[var1-[a,b,..],var2-[c,d..] ..] - [[c,d]/0.3,

simple factor one variable
factor(var1,[a/0.3,b/0.2,c/0.1])

factor(var1,
       [a/factor(var2,[c/0.3,d/0.4]),
	b/factor(var2,[c/0.2,d/0.4])
       ]
      ).
----------------------------------------------------------------------*/

% sums out last dimension of a factor
sum_out(factor(_,List), Num) :-
	\+ List = [_Val/factor(_,_)|_], %just numbers
	!,
	just_add(List,0,Num).
sum_out(factor(Var1,List), factor(Var1,NewList)) :-
	sum_for_each_val(List,NewList).

sum_for_each_val([],[]).
sum_for_each_val([Val/Factor|T],[Val/NewFactor|T2]) :-
	sum_out(Factor,NewFactor),
	sum_for_each_val(T,T2).

just_add([],Num,Num).
just_add([_Val/X|T],In,Out) :-
	Mid is In + X,
	just_add(T,Mid,Out).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%common outer variable
fact_mult(factor(V1,Vals11),factor(V1,Vals12),factor(V1,Vals13)) :-
	!,
	fact_mult_for_each_val(Vals11,Vals12,Vals13).

% V1 is not in 2nd factor
fact_mult(factor(V1,Vals11),Factor,factor(V1,Vals13)) :-
	!,
	expand(Vals11,Factor,Vals13).

% first a number, second a factor
fact_mult(X,factor(V,Vals),factor(V,NewVals)) :-
	!,
	expand(Vals,X,NewVals).

% not factors 2 numbers!
fact_mult(X,Y,Prod) :-
	Prod is X*Y.

expand([],_,[]).
expand([Val/F|T],Factor,[Val/NewF|T2]) :-
	fact_mult(F,Factor,NewF),
	expand(T,Factor,T2).
	
fact_mult_for_each_val([],[],[]).
fact_mult_for_each_val([Val/F1|T1],[Val/F2|T2],[Val/F3|T3]) :-
	fact_mult(F1,F2,F3),
	fact_mult_for_each_val(T1,T2,T3).